import boto3
from boto3.session import Session
from mcp.server.fastmcp import FastMCP
from datetime import date, datetime, timedelta
from typing import Dict, Any, Optional, List
from collections import defaultdict

from aws_finops_mcp_server.helpers.util import (
    profiles_to_use, 
    ApiErrors,
    get_stopped_ec2,
    get_unattached_ebs_volumes,
    get_unassociated_eips,
    get_budget_data,
    cost_filters,
    )

mcp = FastMCP("aws_finops")

@mcp.tool(annotations={"readOnlyHint": True})
async def get_cost(
    profiles: Optional[List[str]] = None, 
    all_profiles: bool = False,
    time_range_days: Optional[int] = None,
    start_date_iso: Optional[str] = None, 
    end_date_iso: Optional[str] = None,   
    tags: Optional[List[str]] = None,
    dimensions: Optional[List[str]] = None,
    group_by: Optional[str] = "SERVICE",
) -> Dict[str, Any]:
    """
    Get cost data for a specified AWS profile for a single defined period.
    The period can be defined by 'time_range_days' (last N days including today)
    OR by explicit 'start_date_iso' and 'end_date_iso'.
    If 'start_date_iso' and 'end_date_iso' are provided, they take precedence.
    If no period is defined, defaults to the current month to date.
    Tags can be provided as a list of "Key=Value" strings to filter costs.
    Dimensions can be provided as a list of "Key=Value" strings to filter costs by specific dimensions.
    If no tags or dimensions are provided, all costs will be returned.
    Grouping can be done by a specific dimension, default is "SERVICE".
    
    Args:
        profile_name: The AWS CLI profile name to use.
        all_profiles: If True, use all available profiles; otherwise, use the specified profiles.
        time_range_days: Optional. Number of days for the cost data (e.g., last 7 days).
        start_date_iso: Optional. The start date of the period (inclusive) in YYYY-MM-DD format.
        end_date_iso: Optional. The end date of the period (inclusive) in YYYY-MM-DD format.
        tags: Optional. List of cost allocation tags (e.g., ["Team=DevOps", "Env=Prod"]).
        dimensions: Optional. List of dimensions to filter costs by (e.g., ["REGION=us-east-1", "AZ=us-east-1a"]).
        group_by: Optional. The dimension to group costs by (default is "SERVICE").
    Returns:
        Dict: Processed cost data for the specified period.
    """
    if all_profiles:
        profiles_to_query, errors_for_profiles = profiles_to_use(all_profiles=True)
        if not profiles_to_query:
            return {"error": "No valid profiles found."}
    else:
        profiles_to_query, errors_for_profiles = profiles_to_use(profiles)
        if not profiles_to_query:
            return {"error": "No valid profiles found."}
    
    cost_data = {}
    
    for account_id, profile in profiles_to_query.items():
        primary_profile = profile[0]

        try:
            session = boto3.Session(profile_name=primary_profile)
            account_id = session.client("sts").get_caller_identity().get("Account")
            ce = session.client("ce")

            cost_explorer_kwargs = cost_filters(tags, dimensions)

            today = date.today()
            period_start_date: date
            period_api_end_date: date # Exclusive end date for Cost Explorer API
            period_display_end_date: date # Inclusive end date for display in results

            if start_date_iso and end_date_iso:
                try:
                    period_start_date = datetime.strptime(start_date_iso, "%Y-%m-%d").date()
                    period_display_end_date = datetime.strptime(end_date_iso, "%Y-%m-%d").date()
                    period_api_end_date = period_display_end_date + timedelta(days=1)
                    if period_start_date > period_display_end_date:
                        return {"profile_name": primary_profile, "status": "error", "message": "Start date cannot be after end date."}
                except ValueError:
                    return {"profile_name": primary_profile, "status": "error", "message": "Invalid date format. Use YYYY-MM-DD."}
            elif time_range_days is not None:
                if time_range_days <= 0:
                    return {"profile_name": primary_profile, "status": "error", "message": "time_range_days must be positive."}
                period_display_end_date = today
                period_start_date = today - timedelta(days=time_range_days - 1) # -1 to make it inclusive of N days
                period_api_end_date = today + timedelta(days=1)
            else: # Default to current month to date
                period_start_date = today.replace(day=1)
                period_display_end_date = today
                period_api_end_date = today + timedelta(days=1)
            
            # Period Total Cost
            total_cost_response = ce.get_cost_and_usage(
                TimePeriod={"Start": period_start_date.isoformat(), "End": period_api_end_date.isoformat()},
                Granularity="MONTHLY" if not time_range_days and not (start_date_iso and end_date_iso) else "DAILY",
                Metrics=["UnblendedCost"],
                **cost_explorer_kwargs,
            )
            period_total_cost = 0.0
            if total_cost_response.get("ResultsByTime"):
                for month_result in total_cost_response["ResultsByTime"]:
                    if "Total" in month_result and \
                    "UnblendedCost" in month_result["Total"] and \
                    "Amount" in month_result["Total"]["UnblendedCost"]:
                        period_total_cost += float(month_result["Total"]["UnblendedCost"]["Amount"])
                
            # Period Cost by Service
            service_granularity = "DAILY"
            if not time_range_days and not (start_date_iso and end_date_iso):
                service_granularity = "MONTHLY"

            cost_by_service_response = ce.get_cost_and_usage(
                TimePeriod={"Start": period_start_date.isoformat(), "End": period_api_end_date.isoformat()},
                Granularity=service_granularity,
                Metrics=["UnblendedCost"],
                GroupBy=[{"Type": "DIMENSION", "Key": group_by}],
                **cost_explorer_kwargs,
            )
            
            aggregated_service_costs: Dict[str, float] = defaultdict(float)
            if cost_by_service_response.get("ResultsByTime"):
                for result_by_time in cost_by_service_response["ResultsByTime"]:
                    for group in result_by_time.get("Groups", []):
                        service = group["Keys"][0]
                        amount = float(group["Metrics"]["UnblendedCost"]["Amount"])
                        aggregated_service_costs[service] += amount
            
            sorted_service_costs = dict(sorted(aggregated_service_costs.items(), key=lambda item: item[1], reverse=True))
            processed_service_costs = {k: round(v, 2) for k, v in sorted_service_costs.items() if v > 0.001}

            cost_data[f"Profile Name: {primary_profile}"] = {
                "AWS Account #": account_id,
                "Period Start Date": period_start_date.isoformat(),
                "Period End Date": period_display_end_date.isoformat(),
                "Total Cost": round(period_total_cost, 2),
                f"Cost By {group_by}": processed_service_costs,
                "Status": "success"
            }

        except Exception as e:
            cost_data[primary_profile] = {
                "profile_name": primary_profile,
                "status": "error",
                "message": str(e)
            } 
    return {"accounts_cost_data": cost_data, "errors_for_profiles": errors_for_profiles}


@mcp.tool(annotations={"readOnlyHint": True})
async def run_finops_audit(
    regions: List[str],
    profiles: Optional[List[str]] = None,
    all_profiles: bool = False,
    ) -> Dict[Any, Any]:

    """
    Get FinOps Audit Report findings for your AWS CLI Profiles.
    Each Audit Report includes:
        Stopped EC2 Instances, 
        Un-attached EBS VOlumes,
        Un-associated EIPs,
        Budget Status for your one or more specified AWS profiles. Except the budget status, other resources are region specific. 

    Args:
        List of AWS CLI profiles as strings.
        List of AWS Regions as strings.
        all_profiles: If True, use all available profiles; otherwise, use the specified profiles.

    Returns:
        Processed Audit data for specified CLI Profile and regions in JSON(dict) format with errors caught from APIs.
    """

    audit_report: Dict[str, Any] = defaultdict(list)

    if all_profiles:
        profiles_to_query, errors_for_profiles = profiles_to_use(all_profiles=True)
        if not profiles_to_query:
            return {"error": "No valid profiles found."}
    else:
        profiles_to_query, errors_for_profiles = profiles_to_use(profiles)
        if not profiles_to_query:
            return {"error": "No valid profiles found."}

    unique_profiles_to_query: List[str] = []
    if profiles_to_query:
        for accountid, profile in profiles_to_query.items():
            unique_profiles_to_query.append(profile[0])
    
    for profile in unique_profiles_to_query:
        session = boto3.Session(profile_name=profile)
        accountid = session.client('sts').get_caller_identity().get('Account')

        stopped_ec2, stoppedEc2Errors = get_stopped_ec2(session, regions)
        unattachedEBSVolumes, errorsGettingVolumes = get_unattached_ebs_volumes(session, regions)
        un_attached_eips, errorsGettingEIPs = get_unassociated_eips(session, regions)
        budget_status, error_getting_budgets = get_budget_data(session, accountid)

        audit_report[f"Profile Name: {profile}"].append({
            "AWS Account": accountid, 
            "Stopped EC2 Instances": stopped_ec2,
            "Unattached EBS Volumes": unattachedEBSVolumes,
            "Un-associated EIPs":  un_attached_eips,
            "Budget Status": budget_status,
            "Errors getting stopped EC2 instances": stoppedEc2Errors,
            "Errors getting EBS volumes": errorsGettingVolumes,
            "Errors getting EIPS": errorsGettingEIPs,
            "Errors getting Budgets": error_getting_budgets})
        
    return {"Audit Report": dict(audit_report),
        "Error processing profiles": errors_for_profiles}

  
def run_server():
    """
    Entry point to the FastMCP server.
    """
    mcp.run(transport='stdio')

if __name__ == "__main__":
    run_server()


import boto3
from typing import List, Optional, Any, Dict, Tuple
from collections import defaultdict
from boto3.session import Session

ApiErrors = Dict[str, str]

def profiles_to_use(
        profiles: Optional[List[str]] = None,
        all_profiles: Optional[bool] = False
    )-> Tuple[Dict[str, List[str]], ApiErrors]:
    """
    Filters a list of AWS profiles, checks if they exist, retrieves their
    AWS Account ID, and groups the profiles by these Account IDs.

    Args:
        profiles: A list of profile names to process.
        all_profiles: If True, retrieves all available profiles from the current session.
                      If False, only processes the provided list of profiles.

    Returns:
        A dictionary where keys are AWS Account IDs (str) and
        values are lists of profile names (List[str]) that belong to that account.
        Profiles that don't exist or cause errors during Account ID retrieval are skipped.
    """

    profile_errors: ApiErrors =  {}
    

    session = boto3.Session()
    available_profiles = session.available_profiles
    account_to_profiles_map: Dict[str, List[str]] = defaultdict(list)

    if all_profiles:
        profiles_to_query = available_profiles
    else:
        profiles_to_query = profiles

    for profile in profiles_to_query:
        if profile in available_profiles:
            try:
                session = boto3.Session(profile_name=profile)
                accountId = session.client('sts').get_caller_identity().get('Account')
                if accountId:
                    account_to_profiles_map[accountId].append(profile)
            except Exception as e:
                profile_errors[f"Error processing profile {profile}"] = str(e)
                continue
        else:
            profile_errors[f"Error processing for profile {profile}"] = f"Profile {profile} does not exist."

    return dict(account_to_profiles_map), profile_errors 
            

def get_accessible_regions(profile: str) -> Tuple[List[str], ApiErrors]:
    """
    Retrieves the list of AWS regions accessible with the given profile.

    Args:
        profile: The AWS CLI profile name to use.

    Returns:
        A list of accessible AWS regions for the specified profile.
    """
    region_errors: ApiErrors = {}
    try:
        session = boto3.Session(profile_name=profile)
        accessible_regions =  session.get_available_regions('ec2')
    except Exception as e:
        region_errors[f"Error retrieving regions for profile {profile}"] = str(e)

    return accessible_regions, region_errors


def get_stopped_ec2(
        session: Session,
        regions: List[str]
    ) -> Tuple[Dict[Any, Any], ApiErrors]: 

    """
    Retrieves stopped EC2 instances for each region using the provided session.

    Args:
        session: A boto3 Session object with a specific profile.
        regions: A list of AWS region names.

    Returns:
        A tuple containing:
        - A dictionary mapping each region to a list of stopped EC2 instance IDs.
        - A dictionary of errors that occurred while querying regions.
    """

    stopped_ec2 = defaultdict(list)
    stopped_ec2_errors: ApiErrors = {}
    for region in regions:
        try:
            ec2client = session.client("ec2", region_name=region)
            paginator = ec2client.get_paginator("describe_instances")

            for page in paginator.paginate(
                Filters=[
                    {"Name": "instance-state-name", "Values": ["stopped"]}
                     ]
            ):
                for reservation in page["Reservations"]:
                    for instance in reservation["Instances"]:
                        stopped_ec2[region].append({
                            "InstanceID" : instance["InstanceId"],
                            "Instance Type": instance["InstanceType"]
                            })
        except Exception as e:
            stopped_ec2_errors[region] = str(e)

    return dict(stopped_ec2), stopped_ec2_errors


def get_unattached_ebs_volumes(
        session: Session,
        regions: List[str]
) -> Tuple[Dict[Any, Any], ApiErrors]:
    
    """
    Get a list of un-attached EBS volumes for specified regions. 

    Args: boto3 session and a list of AWS regions
    """
    
    available_ebs_volumes = defaultdict(list)
    errors_getting_ebs_volumes: ApiErrors = {}
    
    for region in regions:
        try:
            ebsClient = session.client("ec2", region_name=region)
            paginator = ebsClient.get_paginator("describe_volumes")

            for page in paginator.paginate(
                Filters=[
                    {"Name": "status",
                    "Values": ["available"]
                    }
                ]):
                for volume in page["Volumes"]:
                    available_ebs_volumes[region].append({
                        "VolumeId": volume["VolumeId"],
                        "Volume Type": volume["VolumeType"],
                        "Volume Size": f"{volume["Size"]} GB"
                    })
        except Exception as e:
            errors_getting_ebs_volumes[region] = str(e)

    return dict(available_ebs_volumes), errors_getting_ebs_volumes

def get_unassociated_eips(
        session: Session,
        regions: List[str]
    ) -> Tuple[Dict[str, Any], ApiErrors]:

    """
    Get a list of un-associated EIPs for specified regions. 

    Args: boto3 session and a list of AWS regions
    """

    un_associated_eips = defaultdict(list)
    errors_getting_eips: ApiErrors = {}
    
    for region in regions:
        try:
            eipClient = session.client("ec2", region_name=region)
            response = eipClient.describe_addresses()
            for address in response["Addresses"]:
                if "AssociationId" not in address:
                    un_associated_eips[region].append({
                            "Allocation ID": address["AllocationId"],
                            "Public IP": address["PublicIp"]
                        })
        except Exception as e:
            errors_getting_eips[region] = str(e)

    return dict(un_associated_eips), errors_getting_eips

def get_budget_data(
        session: Session,
        accountId: str,
    ) -> Dict[str, str]:

    """
    Get a list of Budgets for specied AWS Account

    Args: boto3 session and accountId
    """

    budget_data: Dict[str, str] = defaultdict(list)
    budgetClient = session.client('budgets', region_name="us-east-1")
    error_getting_budgets: Dict[str, str] = {}

    try:
        budgetClient = session.client('budgets', region_name="us-east-1")
        response = budgetClient.describe_budgets(AccountId=accountId)

        for budget in response.get("Budgets", []):
            name = budget["BudgetName"]
            limit = float(budget["BudgetLimit"]["Amount"])
            actual = float(budget["CalculatedSpend"]["ActualSpend"]["Amount"])
            forecast = float(budget["CalculatedSpend"].get("ForecastedSpend", {}).get("Amount", 0.0))

            if actual > limit:
                status = "Over Budget"
            elif forecast > limit:
                status = "Forecasted to Exceed"
            else:
                status = "Under Budget"

            budget_data[name] = {
                "Limit": f"{limit:.2f}",
                "Actual Spend": f"{actual:.2f}",
                "Forecasted Spend": f"{forecast:.2f}" if forecast else "N/A",
                "Status": status
            }

    except Exception as e:
        error_getting_budgets["Budget Error"] = str(e)

    return budget_data, error_getting_budgets


def cost_filters(
    tags: Optional[List[str]] = None,
    dimensions: Optional[List[str]] = None,
) -> Dict[str, Any]:
    
    """
    Constructs filter parameters for AWS Cost Explorer API call based on provided tags and dimensions.
    """
    
    tag_filters_list: List[Dict[str, Any]] = []
    dimension_filters_list: List[Dict[str, Any]] = []
    filter_param: Optional[Dict[str, Any]] = None
    cost_explorer_kwargs: Dict[str, Any] = {}

    if tags:
        for t_str in tags:
            if "=" in t_str:
                key, value = t_str.split("=", 1)
                tag_filters_list.append({"Key": key, "Values": [value]})

    if dimensions:
        for d_str in dimensions:
            if "=" in d_str:
                key, value = d_str.split("=", 1)
                dimension_filters_list.append({"Key": key, "Values": [value]})

    filters = []
    if tag_filters_list:
        if len(tag_filters_list) == 1:
            filters.append({"Tags": tag_filters_list[0]})
        else:
            filters.append({"And": [{"Tags": f} for f in tag_filters_list]})

    if dimension_filters_list:
        if len(dimension_filters_list) == 1:
            filters.append({"Dimensions": dimension_filters_list[0]})
        else:
            filters.append({"And": [{"Dimensions": f} for f in dimension_filters_list]})

    if len(filters) == 1:
        filter_param = filters[0]
    elif len(filters) > 1:
        filter_param = {"And": filters}

    if filter_param:
        cost_explorer_kwargs["Filter"] = filter_param

    return cost_explorer_kwargs
